import base64
import contextlib
import logging
import urllib.parse
from pathlib import Path
from ssl import SSLContext
from typing import Any, AsyncIterator, Dict, List, Mapping, Optional

import aiohttp
from aiohttp import web
from tenacity import AsyncRetrying, retry_if_exception_type, stop_after_attempt, wait_fixed

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import BinarySize

from .exceptions import (
    GPFSAPIError,
    GPFSInvalidBodyError,
    GPFSJobCancelledError,
    GPFSJobFailedError,
    GPFSNotFoundError,
    GPFSUnauthorizedError,
)
from .types import (
    GPFSDisk,
    GPFSFilesystem,
    GPFSJob,
    GPFSJobStatus,
    GPFSQuota,
    GPFSQuotaType,
    GPFSStoragePoolUsage,
    GPFSSystemHealthState,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


def error_handler(inner):
    async def outer(*args, **kwargs):
        try:
            return await inner(*args, **kwargs)
        except web.HTTPBadRequest:
            raise GPFSInvalidBodyError
        except web.HTTPNotFound:
            raise GPFSNotFoundError

    return outer


class GPFSAPIClient:
    api_endpoint: str
    username: str
    password: str

    ssl: Optional[bool | SSLContext]

    def __init__(
        self, endpoint: str, username: str, password: str, ssl: Optional[bool | SSLContext] = None
    ) -> None:
        self.api_endpoint = endpoint
        self.username = username
        self.password = password
        self.ssl = ssl

    @property
    def _req_header(self) -> Mapping[str, str]:
        credential = base64.b64encode(f"{self.username}:{self.password}".encode()).decode()
        return {
            "Authorization": f"Basic {credential}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

    async def _build_request(
        self,
        sess: aiohttp.ClientSession,
        method: str,
        path: str,
        body: Optional[Any] = None,
    ) -> aiohttp.ClientResponse:
        match method:
            case "GET":
                func = sess.get
            case "POST":
                func = sess.post
            case "PUT":
                func = sess.put
            case "PATCH":
                func = sess.patch
            case "DELETE":
                func = sess.delete
            case _:
                raise GPFSAPIError(f"Unsupported request method {method}")
        try:
            if method == "GET" or method == "DELETE":
                return await func("/scalemgmt/v2" + path, headers=self._req_header, ssl=self.ssl)
            else:
                return await func(
                    "/scalemgmt/v2" + path, headers=self._req_header, json=body, ssl=self.ssl
                )
        except web.HTTPUnauthorized:
            raise GPFSUnauthorizedError

    async def _wait_for_job_done(self, jobs: List[GPFSJob]) -> None:
        for job_to_wait in jobs:
            async for attempt in AsyncRetrying(
                wait=wait_fixed(0.5),
                stop=stop_after_attempt(100),
                retry=retry_if_exception_type(web.HTTPNotFound),
            ):
                job = await self.get_job(job_to_wait.jobId)
                if job.status == GPFSJobStatus.COMPLETED:
                    return
                elif job.status == GPFSJobStatus.FAILED:
                    raise GPFSJobFailedError(job.result.to_json() if job.result is not None else "")
                elif job.status == GPFSJobStatus.CANCELLED:
                    raise GPFSJobCancelledError

    @contextlib.asynccontextmanager
    async def _build_session(self) -> AsyncIterator[aiohttp.ClientSession]:
        async with aiohttp.ClientSession(
            base_url=self.api_endpoint,
        ) as sess:
            yield sess

    @error_handler
    async def get_job(self, job_id: int) -> GPFSJob:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", f"/jobs/{job_id}")
            data = await response.json()
            if len(data.get("jobs", [])) == 0:
                raise GPFSNotFoundError
            return GPFSJob.from_dict(data["jobs"][0])

    @error_handler
    async def list_fs(self) -> List[GPFSFilesystem]:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", "/filesystems")
            data = await response.json()
        return [GPFSFilesystem.from_dict(fs_json) for fs_json in data["filesystems"]]

    @error_handler
    async def get_fs(self, fs_name: str) -> GPFSFilesystem:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", f"/filesystems/{fs_name}")
            data = await response.json()
        if len(data.get("filesystems", [])) == 0:
            raise GPFSNotFoundError
        return GPFSFilesystem.from_dict(data["filesystems"][0])

    @error_handler
    async def list_fs_pools(self, fs_name: str) -> List[GPFSStoragePoolUsage]:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", f"/filesystems/{fs_name}/pools")
            data = await response.json()
        if len(data.get("storagePool", [])) == 0:
            raise GPFSNotFoundError
        return [GPFSStoragePoolUsage.from_dict(x) for x in data["storagePool"]]

    @error_handler
    async def get_fs_pool(self, fs_name: str, pool_name: str) -> GPFSStoragePoolUsage:
        async with self._build_session() as sess:
            response = await self._build_request(
                sess, "GET", f"/filesystems/{fs_name}/pools/{pool_name}"
            )
            data = await response.json()
        if len(data.get("storagePool", [])) == 0:
            raise GPFSNotFoundError
        return GPFSStoragePoolUsage.from_dict(data["storagePool"][0])

    @error_handler
    async def list_fs_disks(self, fs_name: str) -> List[GPFSDisk]:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", f"/filesystems/{fs_name}/disks")
            data = await response.json()
        if len(data.get("disks", [])) == 0:
            raise GPFSNotFoundError
        return [GPFSDisk.from_dict(x) for x in data["disks"]]

    @error_handler
    async def list_quotas(
        self, fs_name: str, quota_type: GPFSQuotaType = GPFSQuotaType.FILESET
    ) -> List[GPFSQuota]:
        async with self._build_session() as sess:
            query = urllib.parse.urlencode({"filter": f"entityType={quota_type}"})
            response = await self._build_request(
                sess,
                "GET",
                f"/filesystems/{fs_name}/quotas?{query}",
            )
            data = await response.json()
        return [GPFSQuota.from_dict(quota_info) for quota_info in data["quotas"]]

    @error_handler
    async def list_fileset_quotas(
        self,
        fs_name: str,
        fileset_name: str,
        quota_type: GPFSQuotaType = GPFSQuotaType.FILESET,
    ) -> List[GPFSQuota]:
        async with self._build_session() as sess:
            query = urllib.parse.urlencode({"filter": f"entityType={quota_type}"})
            response = await self._build_request(
                sess,
                "GET",
                f"/filesystems/{fs_name}/filesets/{fileset_name}/quotas?{query}",
            )
            data = await response.json()
            log.debug("response: {}", data)
        return [GPFSQuota.from_dict(quota_info) for quota_info in data["quotas"]]

    @error_handler
    async def set_quota(
        self,
        fs_name: str,
        fileset_name: str,
        limit_bytes: int,
    ) -> None:
        limit_str = str(limit_bytes)
        body = {
            "operationType": "setQuota",
            "quotaType": GPFSQuotaType.FILESET,
            "objectName": fileset_name,
            "blockSoftLimit": limit_str,
            "blockHardLimit": limit_str,
        }
        async with self._build_session() as sess:
            response = await self._build_request(
                sess,
                "POST",
                f"/filesystems/{fs_name}/quotas",
                body,
            )
            data = await response.json()
            await self._wait_for_job_done([GPFSJob.from_dict(x) for x in data["jobs"]])

    @error_handler
    async def remove_quota(self, fs_name: str, fileset_name: str) -> None:
        await self.set_quota(fs_name, fileset_name, BinarySize(0))

    @error_handler
    async def create_fileset(
        self,
        fs_name: str,
        fileset_name: str,
        path: Optional[Path] = None,
        owner: Optional[str] = None,
        permissions: Optional[int] = None,
        create_directory=True,
    ) -> None:
        body: Dict[str, Any] = {
            "filesetName": fileset_name,
            "createDirectory": create_directory,
        }
        if owner is not None:
            body["owner"] = owner
        if permissions is not None:
            body["permissions"] = permissions
        if path is not None:
            body["path"] = path.as_posix()
        async with self._build_session() as sess:
            response = await self._build_request(
                sess, "POST", f"/filesystems/{fs_name}/filesets", body
            )
            data = await response.json()
            await self._wait_for_job_done([GPFSJob.from_dict(x) for x in data["jobs"]])

    @error_handler
    async def remove_fileset(
        self,
        fs_name: str,
        fileset_name: str,
    ) -> None:
        async with self._build_session() as sess:
            response = await self._build_request(
                sess, "DELETE", f"/filesystems/{fs_name}/filesets/{fileset_name}"
            )
            await response.json()

    @error_handler
    async def copy_folder(
        self,
        source_fs_name: str,
        source_directory: Path,
        target_fs_name: str,
        target_directory: Path,
    ) -> None:
        encoded_source_dir = urllib.parse.urlencode({"": source_directory})[1:]
        body = {
            "targetFilesystem": target_fs_name,
            "targetFileset": "fset1",
            "targetPath": target_directory.as_posix(),
        }
        async with self._build_session() as sess:
            response = await self._build_request(
                sess,
                "PUT",
                f"/filesystems/{source_fs_name}/directoryCopy/{encoded_source_dir}",
                body,
            )
            data = await response.json()
            await self._wait_for_job_done([GPFSJob.from_dict(x) for x in data["jobs"]])

    @error_handler
    async def check_health(self) -> str:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", "/healthcheck")
            return await response.text()

    @error_handler
    async def get_cluster_info(self) -> Mapping[str, Any]:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", "/cluster")
            data = await response.json()
            return data["cluster"]

    @error_handler
    async def get_metric(
        self,
        query: str,
    ) -> Mapping[str, Any]:
        querystring = urllib.parse.urlencode({"query": query})
        async with self._build_session() as sess:
            response = await self._build_request(
                sess,
                "GET",
                f"/perfmon/data?{querystring}",
            )
            data = await response.json()
            return data

    @error_handler
    async def list_nodes(self) -> List[str]:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", "/nodes")
            data = await response.json()
            return [x["adminNodeName"] for x in data["nodes"]]

    @error_handler
    async def get_node_health(self, node_name: str) -> List[GPFSSystemHealthState]:
        async with self._build_session() as sess:
            response = await self._build_request(sess, "GET", f"/nodes/{node_name}/health/states")
            data = await response.json()
            return [GPFSSystemHealthState.from_dict(x) for x in data["states"]]



from typing import Optional


class GPFSError(Exception):
    message: str

    def __init__(self, message: Optional[str] = None, *args):
        super().__init__(message, *args)

        self.message = message or ""

    def __str__(self) -> str:
        return type(self).__name__


class GPFSInitError(GPFSError):
    pass


class GPFSAPIError(GPFSError):
    pass


class GPFSInvalidBodyError(GPFSAPIError):
    pass


class GPFSUnauthorizedError(GPFSAPIError):
    pass


class GPFSNotFoundError(GPFSAPIError):
    pass


class GPFSInternalError(GPFSAPIError):
    pass


class GPFSNoMetricError(GPFSError):
    pass


class GPFSJobFailedError(GPFSError):
    pass


class GPFSJobCancelledError(GPFSError):
    pass



import asyncio
import functools
import grp
import logging
import multiprocessing
import os
import pwd
import ssl
import sys
from contextlib import asynccontextmanager as actxmgr
from pathlib import Path
from pprint import pprint
from typing import Any, AsyncIterator, Sequence

import aiomonitor
import aiotools
import click
from aiohttp import web
from setproctitle import setproctitle

from ai.backend.common.config import (
    ConfigurationError,
    override_key,
    redis_config_iv,
)
from ai.backend.common.defs import REDIS_STREAM_DB
from ai.backend.common.events import EventDispatcher, EventProducer
from ai.backend.common.logging import BraceStyleAdapter, Logger
from ai.backend.common.types import LogSeverity
from ai.backend.common.utils import env_info

from . import __version__ as VERSION
from .config import load_local_config, load_shared_config
from .context import EVENT_DISPATCHER_CONSUMER_GROUP, RootContext
from .watcher import WatcherClient, main_job

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


def _is_root() -> bool:
    return os.geteuid() == 0


@aiotools.server_context
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Sequence[Any],
) -> AsyncIterator[None]:
    setproctitle(f"backend.ai: storage-proxy worker-{pidx}")
    try:
        asyncio.get_child_watcher()
    except (AttributeError, NotImplementedError):
        pass
    log_endpoint = _args[1]
    logger = Logger(_args[0]["logging"], is_master=False, log_endpoint=log_endpoint)
    with logger:
        async with server_main(loop, pidx, _args):
            yield


async def check_migration(ctx: RootContext) -> None:
    from .migration import check_latest

    await check_latest(ctx)


@actxmgr
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    _args: Sequence[Any],
) -> AsyncIterator[None]:
    local_config = _args[0]
    loop.set_debug(local_config["debug"]["asyncio"])
    m = aiomonitor.Monitor(
        loop,
        termui_port=local_config["storage-proxy"]["aiomonitor-termui-port"] + pidx,
        webui_port=local_config["storage-proxy"]["aiomonitor-webui-port"] + pidx,
        console_enabled=False,
        hook_task_factory=local_config["debug"]["enhanced-aiomonitor-task-info"],
    )
    m.prompt = f"monitor (storage-proxy[{pidx}@{os.getpid()}]) >>> "
    m.console_locals["local_config"] = local_config
    aiomon_started = False
    try:
        m.start()
        aiomon_started = True
    except Exception as e:
        log.warning("aiomonitor could not start but skipping this error to continue", exc_info=e)

    try:
        etcd = load_shared_config(local_config)
        try:
            redis_config = redis_config_iv.check(
                await etcd.get_prefix("config/redis"),
            )
            log.info("PID: {0} - configured redis_config: {1}", pidx, redis_config)
        except Exception as e:
            log.exception("Unable to read config from etcd")
            raise e

        event_producer = await EventProducer.new(
            redis_config,
            db=REDIS_STREAM_DB,
            log_events=local_config["debug"]["log-events"],
        )
        log.info("PID: {0} - Event producer created. (redis_config: {1})", pidx, redis_config)
        event_dispatcher = await EventDispatcher.new(
            redis_config,
            db=REDIS_STREAM_DB,
            log_events=local_config["debug"]["log-events"],
            node_id=local_config["storage-proxy"]["node-id"],
            consumer_group=EVENT_DISPATCHER_CONSUMER_GROUP,
        )
        log.info("PID: {0} - Event dispatcher created. (redis_config: {1})", pidx, redis_config)
        if local_config["storage-proxy"]["use-watcher"]:
            if not _is_root():
                raise ValueError(
                    "Storage proxy must be run as root if watcher is enabled. Else, set"
                    " `use-wathcer` to false in your local config file."
                )
            insock_path: str | None = local_config["storage-proxy"]["watcher-insock-path-prefix"]
            outsock_path: str | None = local_config["storage-proxy"]["watcher-outsock-path-prefix"]
            if insock_path is None or outsock_path is None:
                raise ValueError(
                    "Socket path must be not null. Please set valid socket path to"
                    " `watcher-insock-path-prefix` and `watcher-outsock-path-prefix` in your local"
                    " config file."
                )
            watcher_client = WatcherClient(pidx, insock_path, outsock_path)
            await watcher_client.init()
        else:
            watcher_client = None
        ctx = RootContext(
            pid=os.getpid(),
            node_id=local_config["storage-proxy"]["node-id"],
            pidx=pidx,
            local_config=local_config,
            etcd=etcd,
            event_producer=event_producer,
            event_dispatcher=event_dispatcher,
            watcher=watcher_client,
        )
        async with ctx:
            m.console_locals["ctx"] = ctx
            m.console_locals["client_api_app"] = ctx.client_api_app
            m.console_locals["manager_api_app"] = ctx.manager_api_app

            if pidx == 0:
                await check_migration(ctx)

            client_ssl_ctx = None
            manager_ssl_ctx = None
            if local_config["api"]["client"]["ssl-enabled"]:
                client_ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                client_ssl_ctx.load_cert_chain(
                    str(local_config["api"]["client"]["ssl-cert"]),
                    str(local_config["api"]["client"]["ssl-privkey"]),
                )
            if local_config["api"]["manager"]["ssl-enabled"]:
                manager_ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                manager_ssl_ctx.load_cert_chain(
                    str(local_config["api"]["manager"]["ssl-cert"]),
                    str(local_config["api"]["manager"]["ssl-privkey"]),
                )
            client_api_runner = web.AppRunner(ctx.client_api_app)
            manager_api_runner = web.AppRunner(ctx.manager_api_app)
            await client_api_runner.setup()
            await manager_api_runner.setup()
            client_service_addr = local_config["api"]["client"]["service-addr"]
            manager_service_addr = local_config["api"]["manager"]["service-addr"]
            client_api_site = web.TCPSite(
                client_api_runner,
                str(client_service_addr.host),
                client_service_addr.port,
                backlog=1024,
                reuse_port=True,
                ssl_context=client_ssl_ctx,
            )
            manager_api_site = web.TCPSite(
                manager_api_runner,
                str(manager_service_addr.host),
                manager_service_addr.port,
                backlog=1024,
                reuse_port=True,
                ssl_context=manager_ssl_ctx,
            )
            await client_api_site.start()
            await manager_api_site.start()
            if _is_root():
                uid = local_config["storage-proxy"]["user"]
                gid = local_config["storage-proxy"]["group"]
                os.setgroups(
                    [g.gr_gid for g in grp.getgrall() if pwd.getpwuid(uid).pw_name in g.gr_mem],
                )
                os.setgid(gid)
                os.setuid(uid)
                log.info("Changed process uid:gid to {}:{}", uid, gid)
            log.info("Started service.")
            try:
                yield
            finally:
                log.info("Shutting down...")
                await manager_api_runner.cleanup()
                await client_api_runner.cleanup()
                await event_producer.close()
                await event_dispatcher.close()
                if watcher_client is not None:
                    await watcher_client.close()
    finally:
        if aiomon_started:
            m.close()


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    default=None,
    help=(
        "The config file path. "
        "(default: ./storage-proxy.toml and /etc/backend.ai/storage-proxy.toml)"
    ),
)
@click.option(
    "--debug",
    is_flag=True,
    help="Set the logging level to DEBUG",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogSeverity.__members__.keys()], case_sensitive=False),
    default="INFO",
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    cli_ctx: click.Context,
    config_path: Path,
    log_level: str,
    debug: bool = False,
) -> int:
    try:
        local_config = load_local_config(config_path, debug=debug)
    except ConfigurationError:
        raise click.Abort()
    if debug:
        log_level = "DEBUG"
    override_key(local_config, ("debug", "enabled"), log_level == "DEBUG")
    override_key(local_config, ("logging", "level"), log_level.upper())
    override_key(local_config, ("logging", "pkg-ns", "ai.backend"), log_level.upper())

    multiprocessing.set_start_method("spawn")

    if cli_ctx.invoked_subcommand is None:
        local_config["storage-proxy"]["pid-file"].write_text(str(os.getpid()))
        ipc_base_path = local_config["storage-proxy"]["ipc-base-path"]
        log_sockpath = Path(
            ipc_base_path / f"storage-proxy-logger-{os.getpid()}.sock",
        )
        log_sockpath.parent.mkdir(parents=True, exist_ok=True)
        log_endpoint = f"ipc://{log_sockpath}"
        local_config["logging"]["endpoint"] = log_endpoint
        try:
            logger = Logger(
                local_config["logging"],
                is_master=True,
                log_endpoint=log_endpoint,
            )
            with logger:
                setproctitle("backend.ai: storage-proxy")
                log.info("Backend.AI Storage Proxy", VERSION)
                log.info("Runtime: {0}", env_info())
                log.info("Node ID: {0}", local_config["storage-proxy"]["node-id"])
                log_config = logging.getLogger("ai.backend.agent.config")
                if local_config["debug"]["enabled"]:
                    log_config.debug("debug mode enabled.")
                if "debug" in local_config and local_config["debug"]["enabled"]:
                    print("== Storage proxy configuration ==")
                    pprint(local_config)
                if local_config["storage-proxy"]["event-loop"] == "uvloop":
                    import uvloop

                    uvloop.install()
                    log.info("Using uvloop as the event loop backend")
                insock_path_prefix = local_config["storage-proxy"]["watcher-insock-path-prefix"]
                outsock_path_prefix = local_config["storage-proxy"]["watcher-outsock-path-prefix"]
                num_workers = local_config["storage-proxy"]["num-proc"]

                if local_config["storage-proxy"]["use-watcher"]:
                    if not _is_root():
                        raise ValueError(
                            "Storage proxy must be run as root if watcher is enabled. Else, set"
                            " `use-wathcer` to false in your local config file."
                        )
                    insock_path: str | None = local_config["storage-proxy"][
                        "watcher-insock-path-prefix"
                    ]
                    outsock_path: str | None = local_config["storage-proxy"][
                        "watcher-outsock-path-prefix"
                    ]
                    if insock_path is None or outsock_path is None:
                        raise ValueError(
                            "Socket path must be not null. Please set valid socket path to"
                            " `watcher-insock-path-prefix` and `watcher-outsock-path-prefix` in"
                            " your local config file."
                        )
                    extra_procs = tuple(
                        functools.partial(
                            main_job, worker_pidx, insock_path_prefix, outsock_path_prefix
                        )
                        for worker_pidx in range(num_workers)
                    )
                else:
                    extra_procs = tuple()

                aiotools.start_server(
                    server_main_logwrapper,
                    num_workers=num_workers,
                    extra_procs=extra_procs,
                    args=(local_config, log_endpoint),
                )
                log.info("exit.")
        finally:
            if local_config["storage-proxy"]["pid-file"].is_file():
                # check is_file() to prevent deleting /dev/null!
                local_config["storage-proxy"]["pid-file"].unlink()
    return 0


if __name__ == "__main__":
    sys.exit(main())



import asyncio
import csv
import enum
import logging
import os
import re
from contextlib import asynccontextmanager as actxmgr
from dataclasses import dataclass
from io import StringIO
from pathlib import Path
from typing import Any, AsyncIterator, Iterator, Optional, TypedDict
from uuid import UUID

import aiofiles
import asyncpg
import click
import more_itertools
import tqdm
import yarl

from ai.backend.common.config import redis_config_iv
from ai.backend.common.defs import REDIS_STREAM_DB
from ai.backend.common.events import (
    EventDispatcher,
    EventProducer,
)
from ai.backend.common.logging import BraceStyleAdapter, Logger

from .abc import CAP_FAST_SIZE, AbstractVolume
from .config import load_local_config, load_shared_config
from .context import EVENT_DISPATCHER_CONSUMER_GROUP, RootContext
from .types import VFolderID

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


@dataclass
class VolumeUpgradeInfo:
    orig_version: int
    target_version: int
    volume: AbstractVolume


class MigrationFolderInfo(TypedDict):
    volume_id: str
    folder_id: UUID
    quota_scope_id: str
    src_path: Path
    dst_path: Path
    current_size: Optional[int]
    old_quota: Optional[int]


class VFolderMigrationStatus(enum.StrEnum):
    PENDING = "pending"
    FAILED = "failed"
    COMPLETE = "complete"


async def check_latest(ctx: RootContext) -> list[VolumeUpgradeInfo]:
    volumes_to_upgrade: list[VolumeUpgradeInfo] = []
    volume_infos = ctx.list_volumes()
    for name, info in volume_infos.items():
        async with ctx.get_volume(name) as volume:
            version_path = volume.mount_path / "version.txt"
            if version_path.exists():
                version = int(version_path.read_text().strip())
            else:
                version = 2
            match version:
                case 2:
                    log.warning(
                        "{}: Detected an old vfolder structure (v{})",
                        volume.mount_path,
                        version,
                    )
                    volumes_to_upgrade.append(VolumeUpgradeInfo(2, 3, volume))
                case 3:
                    # already the latest version
                    pass
    return volumes_to_upgrade


def path_to_uuid(p: Path) -> UUID:
    return UUID(f"{p.parent.parent.name}{p.parent.name}{p.name}")


@actxmgr
async def connect_database(dsn: str) -> AsyncIterator[asyncpg.Connection]:
    db_dsn = yarl.URL(dsn).with_scheme("postgres")
    conn = await asyncpg.connect(dsn=str(db_dsn))
    yield conn
    await conn.close()


async def upgrade_2_to_3(
    ctx: RootContext,
    volume: AbstractVolume,
    outfile: str,
    report_path: Optional[Path] = None,
    force_scan_folder_size: bool = False,
) -> None:
    assert ctx.dsn is not None
    rx_two_digits_hex = re.compile(r"^[a-f0-9]{2}$")
    rx_rest_digits_hex = re.compile(r"^[a-f0-9]{28}$")
    log.info("upgrading {} ...", volume.mount_path)
    volume_id = os.fsdecode(volume.mount_path)
    scan_folder_size = force_scan_folder_size or (CAP_FAST_SIZE in await volume.get_capabilities())

    def scan_vfolders(root: Path, *, depth: int = 0) -> Iterator[Path]:
        for p in root.iterdir():
            if depth < 2:
                if p.is_dir() and rx_two_digits_hex.search(p.name):
                    yield from scan_vfolders(p, depth=depth + 1)
            else:
                if p.is_dir() and rx_rest_digits_hex.search(p.name):
                    yield p

    targets = [*scan_vfolders(volume.mount_path)]
    created_quota_scopes: set[str] = set()
    migration_informations: list[MigrationFolderInfo] = []
    with tqdm.tqdm(total=len(targets)) as progbar:
        for target_chunk in more_itertools.ichunked(targets, 10):
            folder_ids: list[UUID] = []
            old_quota_map: dict[UUID, Optional[int]] = {}
            quota_scope_map: dict[UUID, str] = {}
            async with connect_database(ctx.dsn) as conn:
                for target in target_chunk:
                    folder_id = path_to_uuid(target)
                    folder_ids.append(folder_id)
                rows = await conn.fetch(
                    """\
                    SELECT "id", "quota_scope_id", "max_size" FROM vfolders
                    WHERE "id" = ANY($1);
                    """,
                    folder_ids,
                )
                for row in rows:
                    old_quota_map[row["id"]] = (
                        row["max_size"] * (2**20) if row["max_size"] else None
                    )
                    quota_scope_map[row["id"]] = row["quota_scope_id"]

                log.info("checking {} ...".format(", ".join(map(str, folder_ids))))

            for folder_id in folder_ids:
                try:
                    quota_scope_id = quota_scope_map[folder_id]
                except KeyError:
                    continue
                progbar.set_description(
                    "inspecting contents of vfolder {}".format(
                        folder_id,
                    )
                )
                orig_vfid = VFolderID(None, folder_id)
                dst_vfid = VFolderID(quota_scope_id, folder_id)
                try:
                    if scan_folder_size:
                        current_size = int(
                            await volume.fsop_model.scan_tree_size(
                                volume.mangle_vfpath(orig_vfid),
                            )
                        )
                    else:
                        current_size = None
                    if quota_scope_id not in created_quota_scopes:
                        created_quota_scopes.add(quota_scope_id)
                    migration_informations.append(
                        {
                            "volume_id": volume_id,
                            "folder_id": folder_id,
                            "quota_scope_id": quota_scope_id,
                            "src_path": volume.mangle_vfpath(orig_vfid),
                            "dst_path": volume.mangle_vfpath(dst_vfid),
                            "current_size": current_size,
                            "old_quota": old_quota_map[folder_id],
                        }
                    )
                except Exception:
                    log.exception("error during migration of vfolder {}", folder_id)
                finally:
                    progbar.update(1)

    script = (
        "#! /bin/sh\n",
        *[f"mkdir -p {m['dst_path'].parent}\n" for m in migration_informations],
        *[f"mv {m['src_path']} {m['dst_path']}\n" for m in migration_informations],
        f"echo 3 > {volume_id}/version.txt\n",
    )
    if outfile == "-":
        print("".join(script))
    else:
        file_suffix = str(volume.mount_path).split("/")[-1]
        async with aiofiles.open(f"{outfile}.{file_suffix}", "w") as fw:
            await fw.writelines(script)
    if report_path:
        in_memory_file = StringIO()
        fieldnames = [
            "volume_id",
            "folder_id",
            "quota_scope_id",
            "current_size",
            "old_quota",
            "src_path",
            "dst_path",
        ]
        writer = csv.DictWriter(in_memory_file, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(migration_informations)

        async with aiofiles.open(report_path, "w") as csvfile:
            await csvfile.write(in_memory_file.getvalue())


upgrade_handlers = {
    3: upgrade_2_to_3,
}


async def check_and_upgrade(
    local_config: dict[str, Any],
    dsn: str,
    outfile: str,
    report_path: Optional[Path] = None,
    force_scan_folder_size: bool = False,
):
    etcd = load_shared_config(local_config)
    redis_config = redis_config_iv.check(
        await etcd.get_prefix("config/redis"),
    )
    event_producer = await EventProducer.new(
        redis_config,
        db=REDIS_STREAM_DB,
        log_events=local_config["debug"]["log-events"],
    )
    event_dispatcher = await EventDispatcher.new(
        redis_config,
        db=REDIS_STREAM_DB,
        log_events=local_config["debug"]["log-events"],
        node_id=local_config["storage-proxy"]["node-id"],
        consumer_group=EVENT_DISPATCHER_CONSUMER_GROUP,
    )
    ctx = RootContext(
        pid=os.getpid(),
        pidx=0,
        node_id=local_config["storage-proxy"]["node-id"],
        local_config=local_config,
        etcd=etcd,
        dsn=dsn,
        event_producer=event_producer,
        event_dispatcher=event_dispatcher,
        watcher=None,
    )

    async with ctx:
        volumes_to_upgrade = await check_latest(ctx)
        for upgrade_info in volumes_to_upgrade:
            handler = upgrade_handlers[upgrade_info.target_version]
            await handler(
                ctx,
                upgrade_info.volume,
                outfile,
                report_path=report_path,
                force_scan_folder_size=force_scan_folder_size,
            )


@click.command()
@click.argument("outfile")
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help=(
        "The config file path. "
        "[default: ./storage-proxy.toml, ~/.config/backend.ai/storage-proxy.toml, "
        "/etc/backend.ai/storage-proxy.toml (uses the first found one)]"
    ),
)
@click.option(
    "--dsn",
    type=str,
    default="postgres://postgres:develove@localhost:8101/backend",
    help="The DSN of the database connection.",
    show_default=True,
)
@click.option(
    "--report-path",
    "--report",
    type=Path,
    default=None,
    help=(
        "If specified, this program creates a text file which contains information about migrated"
        " folders. Generated report file includes ID of the vFolder, along with its current"
        " occupied size and quota set (both in bytes). Calculating size of the folder without any"
        " help from the storage backend takes so much time, and thus disabled by default. Specify"
        " --force-scan-folder-size flag to override this behavior."
    ),
)
@click.option(
    "--force-scan-folder-size",
    is_flag=True,
    help=(
        "Also scan size of the folder residing in FSOp solution without fast folder size scan"
        " (CAP_FAST_SIZE) capability."
        "WARNING: Enabling this option can slow down whole total migration process a lot!"
    ),
)
@click.option(
    "--debug",
    is_flag=True,
    help="This option will soon change to --log-level TEXT option.",
)
def main(
    outfile: str,
    config_path: Optional[Path],
    dsn: str,
    report_path: Optional[Path],
    force_scan_folder_size: bool,
    debug: bool,
) -> None:
    """
    Print migration script to OUTFILE.
    Pass - as OUTFILE to print results to STDOUT.
    """
    local_config = load_local_config(config_path, debug=debug)
    ipc_base_path = local_config["storage-proxy"]["ipc-base-path"]
    log_sockpath = Path(
        ipc_base_path / f"storage-proxy-logger-{os.getpid()}.sock",
    )
    log_sockpath.parent.mkdir(parents=True, exist_ok=True)
    log_endpoint = f"ipc://{log_sockpath}"
    local_config["logging"]["endpoint"] = log_endpoint
    logger = Logger(
        local_config["logging"],
        is_master=True,
        log_endpoint=log_endpoint,
    )
    with logger:
        asyncio.run(
            check_and_upgrade(
                local_config,
                dsn,
                outfile,
                report_path=report_path,
                force_scan_folder_size=force_scan_folder_size,
            )
        )


if __name__ == "__main__":
    main()



import enum
from dataclasses import dataclass
from typing import Any, List, Optional

from dataclasses_json import DataClassJsonMixin


@dataclass
class GPFSFilesystemBlock(DataClassJsonMixin):
    pools: Optional[str]
    disks: Optional[str]
    blockSize: Optional[int]
    metaDataBlockSize: Optional[int]
    indirectBlockSize: Optional[int]
    minFragmentSize: Optional[int]
    inodeSize: Optional[int]
    logfileSize: Optional[int]
    writeCacheThreshold: Optional[int]


@dataclass
class GPFSFilesystemMount(DataClassJsonMixin):
    mountPoint: Optional[str]
    automaticMountOption: Optional[str]
    additionalMountOptions: Optional[str]
    mountPriority: Optional[int]
    driveLetter: Optional[str]
    remoteDeviceName: Optional[str]
    readOnly: Optional[bool]
    nodesMountedReadWrite: Optional[List[str]]
    nodesMountedReadOnly: Optional[List[str]]
    nodesMountedInternally: Optional[List[str]]
    status: Optional[str]


@dataclass
class GPFSFilesystemReplication(DataClassJsonMixin):
    defaultMetadataReplicas: Optional[int]
    maxMetadataReplicas: Optional[int]
    defaultDataReplicas: Optional[int]
    maxDataReplicas: Optional[int]
    strictReplication: Optional[str]
    logReplicas: Optional[int]


@dataclass
class GPFSFilesystemQuota(DataClassJsonMixin):
    quotasAccountingEnabled: Optional[str]
    quotasEnforced: Optional[str]
    defaultQuotasEnabled: Optional[str]
    perfilesetQuotas: Optional[bool]
    filesetdfEnabled: Optional[bool]


@dataclass
class GPFSFilesystemSettings(DataClassJsonMixin):
    blockAllocationType: Optional[str]
    fileLockingSemantics: Optional[str]
    aclSemantics: Optional[str]
    numNodes: Optional[int]
    dmapiEnabled: Optional[bool]
    exactMTime: Optional[bool]
    suppressATime: Optional[str]
    fastEAEnabled: Optional[bool]
    encryption: Optional[bool]
    maxNumberOfInodes: Optional[int]
    is4KAligned: Optional[bool]
    rapidRepairEnabled: Optional[bool]
    stripeMethod: Optional[str]
    stripedLogs: Optional[bool]
    fileAuditLogEnabled: Optional[bool]


@dataclass
class GPFSFileAuditLogConfig(DataClassJsonMixin):
    auditFilesetDeviceName: Optional[str]
    auditFilesetName: Optional[str]
    auditRetention: Optional[int]
    topicGenNum: Optional[int]
    eventTypes: Optional[str]


@dataclass
class GPFSFilesystem(DataClassJsonMixin):
    oid: Optional[int]
    uuid: Optional[str]
    name: Optional[str]
    version: Optional[str]
    type: Optional[str]
    createTime: Optional[str]
    block: Optional[GPFSFilesystemBlock]
    mount: Optional[GPFSFilesystemMount]
    replication: Optional[GPFSFilesystemReplication]
    quota: Optional[GPFSFilesystemQuota]
    settings: Optional[GPFSFilesystemSettings]
    fileAuditLogConfig: Optional[GPFSFileAuditLogConfig]


@dataclass
class GPFSQuota(DataClassJsonMixin):
    quotaId: Optional[int]
    filesystemName: Optional[str]
    filesetName: Optional[str]
    quotaType: Optional[str]
    objectName: Optional[str]
    objectId: Optional[int]
    blockUsage: Optional[int]
    blockQuota: Optional[int]
    blockLimit: Optional[int]
    blockInDoubt: Optional[int]
    blockGrace: Optional[str]
    filesUsage: Optional[int]
    filesQuota: Optional[int]
    filesLimit: Optional[int]
    filesInDoubt: Optional[int]
    filesGrace: Optional[str]
    defaultQuota: Optional[bool]


@dataclass
class GPFSStoragePoolUsage(DataClassJsonMixin):
    storagePoolName: str
    filesystemName: str
    totalDataInKB: Optional[int] = None
    freeDataInKB: Optional[int] = None
    totalMetaInKB: Optional[int] = None
    freeMetaInKB: Optional[int] = None


@dataclass
class GPFSSystemHealthState(DataClassJsonMixin):
    oid: Optional[int]
    component: Optional[str]
    reportingNode: Optional[str]
    activeSince: Optional[str]
    entityType: Optional[str]
    entityName: Optional[str]
    parentName: Optional[str]
    state: Optional[str]
    reasons: Optional[List[str]]


@dataclass
class GPFSDisk(DataClassJsonMixin):
    name: str
    fileSystem: str
    failureGroup: Optional[str] = None
    type: Optional[str] = None
    storagePool: Optional[str] = None
    status: Optional[str] = None
    availability: Optional[str] = None
    quorumDisk: Optional[bool] = None
    remarks: Optional[str] = None
    size: Optional[int] = None
    availableBlocks: Optional[int] = None
    availableFragments: Optional[int] = None
    nsdServers: Optional[str] = None
    nsdVolumeId: Optional[str] = None


@dataclass
class GPFSJobResult(DataClassJsonMixin):
    commands: Optional[List[str]] = None
    progress: Optional[List[str]] = None
    exitCode: Optional[int] = None
    stderr: Optional[List[str]] = None
    stdout: Optional[List[str]] = None


@dataclass
class GPFSJobRequest(DataClassJsonMixin):
    type: Optional[str] = None
    url: Optional[str] = None
    data: Optional[Any] = None


class GPFSJobStatus(str, enum.Enum):
    RUNNING = "RUNNING"
    CANCELLING = "CANCELLING"
    CANCELLED = "CANCELLED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


@dataclass
class GPFSJob(DataClassJsonMixin):
    jobId: int
    status: GPFSJobStatus
    result: Optional[GPFSJobResult] = None
    request: Optional[GPFSJobRequest] = None
    submitted: Optional[str] = None
    completed: Optional[str] = None
    runtime: Optional[int] = None
    pids: Optional[List[int]] = None


class GPFSQuotaType(str, enum.Enum):
    FILESET = "FILESET"
    USER = "USR"
    GROUP = "GRP"



"""
Manager-facing API
"""
from __future__ import annotations

import json
import logging
import os
from contextlib import contextmanager as ctxmgr
from datetime import datetime
from pathlib import Path, PurePosixPath
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Awaitable,
    Callable,
    Iterator,
    List,
    NotRequired,
    TypedDict,
    cast,
)

import attr
import jwt
import trafaret as t
from aiohttp import hdrs, web

from ai.backend.common import validators as tx
from ai.backend.common.events import (
    DoVolumeMountEvent,
    DoVolumeUnmountEvent,
    VolumeMountableNodeType,
    VolumeMounted,
    VolumeUnmounted,
)
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import AgentId, BinarySize, QuotaScopeID
from ai.backend.storage.exception import ExecutionError
from ai.backend.storage.watcher import ChownTask, MountTask, UmountTask

from .. import __version__
from ..exception import (
    InvalidQuotaConfig,
    InvalidSubpathError,
    QuotaScopeAlreadyExists,
    QuotaScopeNotFoundError,
    StorageProxyError,
    VFolderNotFoundError,
)
from ..types import QuotaConfig, VFolderID
from ..utils import check_params, log_manager_api_entry

if TYPE_CHECKING:
    from ..abc import AbstractVolume
    from ..context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


@web.middleware
async def token_auth_middleware(
    request: web.Request,
    handler: Callable[[web.Request], Awaitable[web.StreamResponse]],
) -> web.StreamResponse:
    skip_token_auth = getattr(handler, "skip_token_auth", False)
    if not skip_token_auth:
        token = request.headers.get("X-BackendAI-Storage-Auth-Token", None)
        if not token:
            raise web.HTTPForbidden()
        ctx: RootContext = request.app["ctx"]
        if token != ctx.local_config["api"]["manager"]["secret"]:
            raise web.HTTPForbidden()
    return await handler(request)


def skip_token_auth(
    handler: Callable[[web.Request], Awaitable[web.StreamResponse]]
) -> Callable[[web.Request], Awaitable[web.StreamResponse]]:
    setattr(handler, "skip_token_auth", True)
    return handler


@skip_token_auth
async def check_status(request: web.Request) -> web.Response:
    async with check_params(request, None) as params:
        await log_manager_api_entry(log, "get_status", params)
        return web.json_response(
            {
                "status": "ok",
                "type": "maanger-facing",
                "storage-proxy": __version__,
            },
        )


@ctxmgr
def handle_fs_errors(
    volume: AbstractVolume,
    vfid: VFolderID,
) -> Iterator[None]:
    try:
        yield
    except OSError as e:
        related_paths = []
        msg = str(e) if e.strerror is None else e.strerror
        if e.filename:
            related_paths.append(str(volume.strip_vfpath(vfid, Path(e.filename))))
        if e.filename2:
            related_paths.append(str(volume.strip_vfpath(vfid, Path(e.filename2))))
        raise web.HTTPBadRequest(
            body=json.dumps(
                {
                    "msg": msg,
                    "errno": e.errno,
                    "paths": related_paths,
                },
            ),
            content_type="application/json",
        )


async def get_volumes(request: web.Request) -> web.Response:
    async def _get_caps(ctx: RootContext, volume_name: str) -> List[str]:
        async with ctx.get_volume(volume_name) as volume:
            return [*await volume.get_capabilities()]

    async with check_params(request, None) as params:
        await log_manager_api_entry(log, "get_volumes", params)
        ctx: RootContext = request.app["ctx"]
        volumes = ctx.list_volumes()
        return web.json_response(
            {
                "volumes": [
                    {
                        "name": name,
                        "backend": info.backend,
                        "path": str(info.path),
                        "fsprefix": str(info.fsprefix),
                        "capabilities": await _get_caps(ctx, name),
                    }
                    for name, info in volumes.items()
                ],
            },
        )


async def get_hwinfo(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "get_hwinfo", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            data = await volume.get_hwinfo()
            return web.json_response(data)


async def create_quota_scope(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        qsid: QuotaScopeID
        options: QuotaConfig | None
        extra_args: NotRequired[dict[str, Any]]

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("qsid"): tx.QuotaScopeID(),
                    t.Key("options", default=None): t.Null | QuotaConfig.as_trafaret(),
                    t.Key("extra_args", default=None): t.Null | t.Mapping(t.String, t.Any),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "create_quota_scope", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            try:
                await volume.quota_model.create_quota_scope(
                    params["qsid"], params["options"], params["extra_args"]
                )
            except QuotaScopeAlreadyExists:
                return web.json_response(
                    {"msg": "Volume already exists with given quota scope."},
                    status=409,
                )
            return web.Response(status=204)


async def get_quota_scope(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        qsid: QuotaScopeID

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("qsid"): tx.QuotaScopeID(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "get_quota_scope", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            quota_usage = await volume.quota_model.describe_quota_scope(params["qsid"])
            if not quota_usage:
                raise QuotaScopeNotFoundError
            return web.json_response(
                {
                    "used_bytes": quota_usage.used_bytes if quota_usage.used_bytes >= 0 else None,
                    "limit_bytes": (
                        quota_usage.limit_bytes if quota_usage.limit_bytes >= 0 else None
                    ),
                }
            )


async def update_quota_scope(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        qsid: QuotaScopeID
        options: QuotaConfig

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("qsid"): tx.QuotaScopeID(),
                    t.Key("options"): QuotaConfig.as_trafaret(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "update_quota_scope", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            quota_usage = await volume.quota_model.describe_quota_scope(params["qsid"])
            if not quota_usage:
                await volume.quota_model.create_quota_scope(params["qsid"], params["options"])
            try:
                await volume.quota_model.update_quota_scope(params["qsid"], params["options"])
            except InvalidQuotaConfig:
                return web.json_response(
                    {"msg": "Invalid quota config option"},
                    status=400,
                )
            return web.Response(status=204)


async def unset_quota(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        qsid: QuotaScopeID

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("qsid"): tx.QuotaScopeID(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "unset_quota", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            quota_usage = await volume.quota_model.describe_quota_scope(params["qsid"])
            if not quota_usage:
                raise QuotaScopeNotFoundError
            await volume.quota_model.unset_quota(params["qsid"])
            return web.Response(status=204)


async def create_vfolder(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        options: dict[str, Any] | None  # deprecated

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("options", default=None): t.Null | t.Dict().allow_extra("*"),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "create_vfolder", params)
        assert params["vfid"].quota_scope_id is not None
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            try:
                await volume.create_vfolder(params["vfid"])
            except QuotaScopeNotFoundError:
                assert params["vfid"].quota_scope_id
                if initial_max_size_for_quota_scope := (params["options"] or {}).get(
                    "initial_max_size_for_quota_scope"
                ):
                    options = QuotaConfig(initial_max_size_for_quota_scope)
                else:
                    options = None
                await volume.quota_model.create_quota_scope(
                    params["vfid"].quota_scope_id, options=options
                )
                await volume.create_vfolder(params["vfid"])
            return web.Response(status=204)


async def delete_vfolder(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "delete_vfolder", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            await volume.delete_vfolder(params["vfid"])
            return web.Response(status=204)


async def clone_vfolder(request: web.Request) -> web.Response:
    class Params(TypedDict):
        src_volume: str
        src_vfid: VFolderID
        dst_volume: str | None  # deprecated
        dst_vfid: VFolderID
        options: dict[str, Any] | None  # deprecated

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("src_volume"): t.String(),
                    t.Key("src_vfid"): tx.VFolderID(),
                    t.Key("dst_volume"): t.String() | t.Null,
                    t.Key("dst_vfid"): tx.VFolderID(),
                    t.Key("options", default=None): t.Null | t.Dict().allow_extra("*"),
                },
            ),
        ),
    ) as params:
        if params["dst_volume"] is not None and params["dst_volume"] != params["src_volume"]:
            raise StorageProxyError("Cross-volume vfolder cloning is not implemented yet")
        await log_manager_api_entry(log, "clone_vfolder", params)
        ctx: RootContext = request.app["ctx"]
        if params["dst_volume"] is not None and params["dst_volume"] != params["src_volume"]:
            raise StorageProxyError("Cross-volume vfolder cloning is not implemented yet")
        async with ctx.get_volume(params["src_volume"]) as src_volume:
            await src_volume.clone_vfolder(
                params["src_vfid"],
                params["dst_vfid"],
            )
        return web.Response(status=204)


async def get_vfolder_mount(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        subpath: str

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("subpath", default="."): t.String(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "get_container_mount", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            try:
                mount_path = await volume.get_vfolder_mount(
                    params["vfid"],
                    params["subpath"],
                )
            except VFolderNotFoundError:
                raise web.HTTPBadRequest(
                    body=json.dumps(
                        {
                            "msg": "VFolder not found",
                            "vfid": str(params["vfid"]),
                        },
                    ),
                    content_type="application/json",
                )
            except InvalidSubpathError as e:
                raise web.HTTPBadRequest(
                    body=json.dumps(
                        {
                            "msg": "Invalid vfolder subpath",
                            "vfid": str(params["vfid"]),
                            "subpath": str(e.args[1]),
                        },
                    ),
                    content_type="application/json",
                )
            return web.json_response(
                {
                    "path": str(mount_path),
                },
            )


async def get_performance_metric(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "get_performance_metric", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            metric = await volume.get_performance_metric()
            return web.json_response(
                {
                    "metric": attr.asdict(cast(attr.AttrsInstance, metric)),
                },
            )


async def fetch_file(request: web.Request) -> web.StreamResponse:
    """
    Direct file streaming API for internal use, such as retrieving
    task logs from a user vfolder ".logs".
    """

    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        relpath: PurePosixPath

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("relpath"): tx.PurePath(relative_only=True),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "fetch_file", params)
        ctx: RootContext = request.app["ctx"]
        response = web.StreamResponse(status=200)
        response.headers[hdrs.CONTENT_TYPE] = "application/octet-stream"
        prepared = False
        try:
            async with ctx.get_volume(params["volume"]) as volume:
                with handle_fs_errors(volume, params["vfid"]):
                    async for chunk in volume.read_file(
                        params["vfid"],
                        params["relpath"],
                    ):
                        if not chunk:
                            return response
                        if not prepared:
                            await response.prepare(request)
                            prepared = True
                        await response.write(chunk)
        except FileNotFoundError:
            response = web.Response(status=404, reason="Log data not found")
        finally:
            if prepared:
                await response.write_eof()
            return response


async def get_metadata(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "get_metadata", params)
        return web.json_response(
            {
                "status": "ok",
            },
        )


async def set_metadata(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        payload: bytes

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("payload"): t.Bytes(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "set_metadata", params)
        return web.json_response(
            {
                "status": "ok",
            },
        )


async def get_vfolder_fs_usage(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "get_vfolder_fs_usage", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            fs_usage = await volume.get_fs_usage()
            return web.json_response(
                {
                    "capacity_bytes": fs_usage.capacity_bytes,
                    "used_bytes": fs_usage.used_bytes,
                },
            )


async def get_vfolder_usage(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                },
            ),
        ),
    ) as params:
        try:
            await log_manager_api_entry(log, "get_vfolder_usage", params)
            ctx: RootContext = request.app["ctx"]
            async with ctx.get_volume(params["volume"]) as volume:
                usage = await volume.get_usage(params["vfid"])
                return web.json_response(
                    {
                        "file_count": usage.file_count,
                        "used_bytes": usage.used_bytes,
                    },
                )
        except ExecutionError:
            return web.Response(
                status=500,
                reason="Storage server is busy. Please try again",
            )


async def get_vfolder_used_bytes(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                },
            ),
        ),
    ) as params:
        try:
            await log_manager_api_entry(log, "get_vfolder_used_bytes", params)
            ctx: RootContext = request.app["ctx"]
            async with ctx.get_volume(params["volume"]) as volume:
                usage = await volume.get_used_bytes(params["vfid"])
                return web.json_response(
                    {
                        "used_bytes": usage,
                    },
                )
        except ExecutionError:
            return web.Response(
                status=500,
                reason="Storage server is busy. Please try again",
            )


async def get_quota(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid", default=None): t.Null | tx.VFolderID,
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "get_quota", params)
        return web.Response(
            status=400,
            reason="get_quota (for individual vfolder) is a deprecated API",
        )


async def set_quota(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        size_bytes: BinarySize

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid", default=None): t.Null | tx.VFolderID,
                    t.Key("size_bytes"): tx.BinarySize,
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "update_quota", params)
        return web.Response(
            status=400,
            reason="set_quota (for individual vfolder) is a deprecated API",
        )


async def mkdir(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        relpath: PurePosixPath
        parents: bool
        exist_ok: bool

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("relpath"): tx.PurePath(relative_only=True),
                    t.Key("parents", default=True): t.ToBool,
                    t.Key("exist_ok", default=False): t.ToBool,
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "mkdir", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            with handle_fs_errors(volume, params["vfid"]):
                await volume.mkdir(
                    params["vfid"],
                    params["relpath"],
                    parents=params["parents"],
                    exist_ok=params["exist_ok"],
                )
        return web.Response(status=204)


async def list_files(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        relpath: PurePosixPath

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("relpath"): tx.PurePath(relative_only=True),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "list_files", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            with handle_fs_errors(volume, params["vfid"]):
                items = [
                    {
                        "name": item.name,
                        "type": item.type.name,
                        "stat": {
                            "mode": item.stat.mode,
                            "size": item.stat.size,
                            "created": item.stat.created.isoformat(),
                            "modified": item.stat.modified.isoformat(),
                        },
                        "symlink_target": item.symlink_target,
                    }
                    async for item in volume.scandir(
                        params["vfid"],
                        params["relpath"],
                        recursive=False,
                    )
                ]
        return web.json_response(
            {
                "items": items,
            },
        )


async def rename_file(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        relpath: PurePosixPath
        new_name: str
        is_dir: bool

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("relpath"): tx.PurePath(relative_only=True),
                    t.Key("new_name"): t.String(),
                    t.Key("is_dir", default=False): t.ToBool,  # ignored since 22.03
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "rename_file", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            with handle_fs_errors(volume, params["vfid"]):
                await volume.move_file(
                    params["vfid"],
                    params["relpath"],
                    params["relpath"].with_name(params["new_name"]),
                )
        return web.Response(status=204)


async def move_file(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        src_relpath: PurePosixPath
        dst_relpath: PurePosixPath

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("src_relpath"): tx.PurePath(relative_only=True),
                    t.Key("dst_relpath"): tx.PurePath(relative_only=True),
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "move_file", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            with handle_fs_errors(volume, params["vfid"]):
                await volume.move_file(
                    params["vfid"],
                    params["src_relpath"],
                    params["dst_relpath"],
                )
        return web.Response(status=204)


async def create_download_session(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        relpath: PurePosixPath
        archive: bool
        unmanaged_path: str | None

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("relpath"): tx.PurePath(relative_only=True),
                    t.Key("archive", default=False): t.ToBool,
                    t.Key("unmanaged_path", default=None): t.Null | t.String,
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "create_download_session", params)
        ctx: RootContext = request.app["ctx"]
        token_data = {
            "op": "download",
            "volume": params["volume"],
            "vfid": str(params["vfid"]),
            "relpath": str(params["relpath"]),
            "exp": datetime.utcnow() + ctx.local_config["storage-proxy"]["session-expire"],
        }
        token = jwt.encode(
            token_data,
            ctx.local_config["storage-proxy"]["secret"],
            algorithm="HS256",
        )
        return web.json_response(
            {
                "token": token,
            },
        )


async def create_upload_session(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        relpath: PurePosixPath
        size: int

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("relpath"): tx.PurePath(relative_only=True),
                    t.Key("size"): t.ToInt,
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "create_upload_session", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            session_id = await volume.prepare_upload(params["vfid"])
        token_data = {
            "op": "upload",
            "volume": params["volume"],
            "vfid": str(params["vfid"]),
            "relpath": str(params["relpath"]),
            "size": params["size"],
            "session": session_id,
            "exp": datetime.utcnow() + ctx.local_config["storage-proxy"]["session-expire"],
        }
        token = jwt.encode(
            token_data,
            ctx.local_config["storage-proxy"]["secret"],
            algorithm="HS256",
        )
        return web.json_response(
            {
                "token": token,
            },
        )


async def delete_files(request: web.Request) -> web.Response:
    class Params(TypedDict):
        volume: str
        vfid: VFolderID
        relpaths: list[PurePosixPath]
        recursive: bool

    async with cast(
        AsyncContextManager[Params],
        check_params(
            request,
            t.Dict(
                {
                    t.Key("volume"): t.String(),
                    t.Key("vfid"): tx.VFolderID(),
                    t.Key("relpaths"): t.List(tx.PurePath(relative_only=True)),
                    t.Key("recursive", default=False): t.ToBool,
                },
            ),
        ),
    ) as params:
        await log_manager_api_entry(log, "delete_files", params)
        ctx: RootContext = request.app["ctx"]
        async with ctx.get_volume(params["volume"]) as volume:
            with handle_fs_errors(volume, params["vfid"]):
                await volume.delete_files(
                    params["vfid"],
                    params["relpaths"],
                    params["recursive"],
                )
        return web.json_response(
            {
                "status": "ok",
            },
        )


async def init_manager_app(ctx: RootContext) -> web.Application:
    app = web.Application(
        middlewares=[
            token_auth_middleware,
        ],
    )
    app["ctx"] = ctx
    app.router.add_route("GET", "/", check_status)
    app.router.add_route("GET", "/status", check_status)
    app.router.add_route("GET", "/volumes", get_volumes)
    app.router.add_route("GET", "/volume/hwinfo", get_hwinfo)
    app.router.add_route("POST", "/quota-scope", create_quota_scope)
    app.router.add_route("GET", "/quota-scope", get_quota_scope)
    app.router.add_route("PATCH", "/quota-scope", update_quota_scope)
    app.router.add_route("DELETE", "/quota-scope/quota", unset_quota)
    app.router.add_route("POST", "/folder/create", create_vfolder)
    app.router.add_route("POST", "/folder/delete", delete_vfolder)
    app.router.add_route("POST", "/folder/clone", clone_vfolder)
    app.router.add_route("GET", "/folder/mount", get_vfolder_mount)
    app.router.add_route("GET", "/volume/performance-metric", get_performance_metric)
    app.router.add_route("GET", "/folder/metadata", get_metadata)
    app.router.add_route("POST", "/folder/metadata", set_metadata)
    app.router.add_route("GET", "/volume/quota", get_quota)
    app.router.add_route("PATCH", "/volume/quota", set_quota)
    app.router.add_route("GET", "/folder/usage", get_vfolder_usage)
    app.router.add_route("GET", "/folder/used-bytes", get_vfolder_used_bytes)
    app.router.add_route("GET", "/folder/fs-usage", get_vfolder_fs_usage)
    app.router.add_route("POST", "/folder/file/mkdir", mkdir)
    app.router.add_route("POST", "/folder/file/list", list_files)
    app.router.add_route("POST", "/folder/file/rename", rename_file)
    app.router.add_route("POST", "/folder/file/move", move_file)
    app.router.add_route("POST", "/folder/file/fetch", fetch_file)
    app.router.add_route("POST", "/folder/file/download", create_download_session)
    app.router.add_route("POST", "/folder/file/upload", create_upload_session)
    app.router.add_route("POST", "/folder/file/delete", delete_files)

    # passive events
    evd = ctx.event_dispatcher
    evd.subscribe(DoVolumeMountEvent, ctx, handle_volume_mount, name="storage.volume.mount")
    evd.subscribe(DoVolumeUnmountEvent, ctx, handle_volume_umount, name="storage.volume.umount")
    return app


async def handle_volume_mount(
    context: RootContext,
    source: AgentId,
    event: DoVolumeMountEvent,
) -> None:
    if context.pidx != 0:
        return
    if context.watcher is None:
        log.debug(f"{context.node_id}: Watcher is disabled. Skip handling mount event.")
        return
    err_msg: str | None = None
    mount_prefix = await context.etcd.get("volumes/_mount")
    volume_mount_path = str(context.local_config["volume"][event.volume_backend_name]["path"])
    mount_path = Path(volume_mount_path, event.dir_name)
    mount_task = MountTask.from_event(event, mount_path=mount_path, mount_prefix=mount_prefix)
    resp = await context.watcher.request_task(mount_task)
    if not resp.succeeded:
        # Produce volume mounted event with error message.
        # And skip chown.
        err_msg = resp.body
        await context.event_producer.produce_event(
            VolumeMounted(
                str(context.node_id),
                VolumeMountableNodeType.STORAGE_PROXY,
                str(mount_path),
                event.quota_scope_id,
                err_msg,
            )
        )
        return

    # change owner of mounted directory
    chown_task = ChownTask(str(mount_path), os.getuid(), os.getgid())
    resp = await context.watcher.request_task(chown_task)
    if not resp.succeeded:
        err_msg = resp.body
    await context.event_producer.produce_event(
        VolumeMounted(
            str(context.node_id),
            VolumeMountableNodeType.STORAGE_PROXY,
            str(mount_path),
            event.quota_scope_id,
            err_msg,
        )
    )


async def handle_volume_umount(
    context: RootContext,
    source: AgentId,
    event: DoVolumeUnmountEvent,
) -> None:
    if context.pidx != 0:
        return
    if context.watcher is None:
        log.debug(f"{context.node_id}: Watcher is disabled. Skip handling umount event.")
        return
    mount_prefix = await context.etcd.get("volumes/_mount")
    timeout = await context.etcd.get("config/watcher/file-io-timeout")
    volume_mount_path = str(context.local_config["volume"][event.volume_backend_name]["path"])
    mount_path = Path(volume_mount_path, event.dir_name)
    umount_task = UmountTask.from_event(
        event,
        mount_path=mount_path,
        mount_prefix=mount_prefix,
        timeout=float(timeout) if timeout is not None else None,
    )
    resp = await context.watcher.request_task(umount_task)
    err_msg = resp.body if not resp.succeeded else None
    if resp.body:
        log.warning(resp.body)
    await context.event_producer.produce_event(
        VolumeUnmounted(
            str(context.node_id),
            VolumeMountableNodeType.STORAGE_PROXY,
            str(mount_path),
            event.quota_scope_id,
            err_msg,
        )
    )



from __future__ import annotations

import asyncio
import contextlib
import json
import os
from pathlib import Path
from subprocess import CalledProcessError
from typing import AsyncIterator, FrozenSet

from ai.backend.common.types import BinarySize, HardwareMetadata

from ..abc import CAP_FAST_FS_SIZE, CAP_FAST_SCAN, CAP_METRIC, CAP_VFOLDER, AbstractFSOpModel
from ..subproc import run
from ..types import CapacityUsage, DirEntry, DirEntryType, FSPerfMetric, Stat, TreeUsage
from ..utils import fstime2datetime
from ..vfs import BaseFSOpModel, BaseVolume
from .purity import PurityClient


class RapidFileToolsFSOpModel(BaseFSOpModel):
    async def copy_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        extra_opts: list[bytes] = []
        if src_path.is_dir():
            extra_opts.append(b"-r")
        try:
            await run(
                [
                    b"pcp",
                    *extra_opts,
                    b"-p",
                    # os.fsencode(src_path / "."),  # TODO: check if "/." is necessary?
                    os.fsencode(src_path),
                    os.fsencode(dst_path),
                ]
            )
        except CalledProcessError as e:
            raise RuntimeError(f'"pcp" command failed: {e.stderr}')

    async def delete_tree(
        self,
        path: Path,
    ) -> None:
        try:
            await run(
                [
                    b"prm",
                    b"-r",
                    os.fsencode(path),
                ]
            )
        except CalledProcessError as e:
            raise RuntimeError(f"'prm' command failed: {e.stderr}")

    def scan_tree(
        self,
        target_path: Path,
        recursive=False,
    ) -> AsyncIterator[DirEntry]:
        raw_target_path = os.fsencode(target_path)

        async def _aiter() -> AsyncIterator[DirEntry]:
            proc = await asyncio.create_subprocess_exec(
                b"pls",
                b"--json",
                raw_target_path,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
            assert proc.stdout is not None
            try:
                while True:
                    line = await proc.stdout.readline()
                    if not line:
                        break
                    line = line.rstrip(b"\n")
                    item = json.loads(line)
                    item_path = Path(item["path"])
                    entry_type = DirEntryType.FILE
                    if item["filetype"] == 40000:
                        entry_type = DirEntryType.DIRECTORY
                    if item["filetype"] == 120000:
                        entry_type = DirEntryType.SYMLINK
                    yield DirEntry(
                        name=item_path.name,
                        path=item_path,
                        type=entry_type,
                        stat=Stat(
                            size=item["size"],
                            owner=str(item["uid"]),
                            # The integer represents the octal number in decimal
                            # (e.g., 644 which actually means 0o644)
                            mode=int(str(item["mode"]), 8),
                            modified=fstime2datetime(item["mtime"]),
                            created=fstime2datetime(item["ctime"]),
                        ),
                        symlink_target="",  # TODO: should be tested on PureStorage
                    )
            finally:
                await proc.wait()

        return _aiter()

    async def scan_tree_usage(
        self,
        path: Path,
    ) -> TreeUsage:
        total_size = 0
        total_count = 0
        raw_target_path = os.fsencode(path)
        # Measure the exact file sizes and bytes
        proc = await asyncio.create_subprocess_exec(
            b"pdu",
            b"-0",
            b"-b",
            b"-a",
            b"-s",
            raw_target_path,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
        assert proc.stdout is not None
        try:
            # TODO: check slowdowns when there are millions of files
            while True:
                try:
                    line = await proc.stdout.readuntil(b"\0")
                    line = line.rstrip(b"\0")
                except asyncio.IncompleteReadError:
                    break
                size, name = line.split(maxsplit=1)
                if len(name) != len(raw_target_path) and name != raw_target_path:
                    total_size += int(size)
                    total_count += 1
        finally:
            await proc.wait()
        return TreeUsage(file_count=total_count, used_bytes=total_size)

    async def scan_tree_size(
        self,
        path: Path,
    ) -> BinarySize:
        proc = await asyncio.create_subprocess_exec(
            b"pdu",
            b"-hs",
            bytes(path),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await proc.communicate()
        if proc.returncode != 0:
            raise RuntimeError(f"pdu command failed: {stderr.decode()}")
        used_bytes, _ = stdout.decode().split()
        return BinarySize.finite_from_str(used_bytes)


class FlashBladeVolume(BaseVolume):
    name = "purestorage"

    async def create_fsop_model(self) -> AbstractFSOpModel:
        return RapidFileToolsFSOpModel(
            self.mount_path,
            self.local_config["storage-proxy"]["scandir-limit"],
        )

    async def init(self) -> None:
        available = True
        try:
            proc = await asyncio.create_subprocess_exec(
                b"pdu",
                b"--version",
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
        except FileNotFoundError:
            available = False
        else:
            try:
                stdout, stderr = await proc.communicate()
                if b"RapidFile Toolkit" not in stdout or proc.returncode != 0:
                    available = False
            finally:
                await proc.wait()
        if not available:
            raise RuntimeError(
                "PureStorage RapidFile Toolkit is not installed. "
                "You cannot use the PureStorage backend for the stroage proxy.",
            )
        self.purity_client = PurityClient(
            self.config["purity_endpoint"],
            self.config["purity_api_token"],
            api_version=self.config["purity_api_version"],
        )
        await super().init()

    async def shutdown(self) -> None:
        await self.purity_client.aclose()

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset(
            [
                CAP_FAST_FS_SIZE,
                CAP_VFOLDER,
                CAP_METRIC,
                CAP_FAST_SCAN,
            ],
        )

    async def get_hwinfo(self) -> HardwareMetadata:
        async with self.purity_client as client:
            metadata = await client.get_metadata()
        return {
            "status": "healthy",
            "status_info": None,
            "metadata": {
                **metadata,
            },
        }

    async def get_fs_usage(self) -> CapacityUsage:
        async with self.purity_client as client:
            usage = await client.get_usage(self.config["purity_fs_name"])
        return CapacityUsage(
            capacity_bytes=usage["capacity_bytes"],
            used_bytes=usage["used_bytes"],
        )

    async def get_performance_metric(self) -> FSPerfMetric:
        async with self.purity_client as client:
            async with contextlib.aclosing(
                client.get_nfs_metric(self.config["purity_fs_name"]),
            ) as items:
                async for item in items:
                    return FSPerfMetric(
                        iops_read=item["reads_per_sec"],
                        iops_write=item["writes_per_sec"],
                        io_bytes_read=item["read_bytes_per_sec"],
                        io_bytes_write=item["write_bytes_per_sec"],
                        io_usec_read=item["usec_per_read_op"],
                        io_usec_write=item["usec_per_write_op"],
                    )
                else:
                    raise RuntimeError(
                        "no metric found for the configured flashblade filesystem",
                    )



from __future__ import annotations

import asyncio
import csv
import logging
import os
import shlex
import subprocess
import time
from collections.abc import AsyncIterator
from contextlib import aclosing
from pathlib import Path
from typing import (
    Any,
    FrozenSet,
    Optional,
)

import aiofiles
import aiofiles.os

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import BinarySize, HardwareMetadata, QuotaScopeID

from ..abc import (
    CAP_FAST_FS_SIZE,
    CAP_FAST_SIZE,
    CAP_METRIC,
    CAP_QUOTA,
    CAP_VFOLDER,
    AbstractFSOpModel,
    AbstractQuotaModel,
)
from ..exception import ExecutionError, InvalidQuotaScopeError, NotEmptyError
from ..subproc import spawn_and_watch
from ..types import (
    SENTINEL,
    CapacityUsage,
    DirEntry,
    DirEntryType,
    FSPerfMetric,
    QuotaConfig,
    QuotaUsage,
    Sentinel,
    Stat,
    TreeUsage,
)
from ..utils import fstime2datetime
from ..vfs import BaseFSOpModel, BaseQuotaModel, BaseVolume
from .netappclient import NetAppClient, StorageID, VolumeID

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]
xcp_lic_check_path = Path("/tmp/backend.ai/storage.netapp.xcp-license-check")


class QTreeQuotaModel(BaseQuotaModel):
    """
    Implements the quota scope model using NetApp's QTrees.
    """

    def __init__(
        self,
        mount_path: Path,
        netapp_client: NetAppClient,
        svm_id: StorageID,
        volume_id: VolumeID,
    ) -> None:
        super().__init__(mount_path)
        self.netapp_client = netapp_client
        self.svm_id = svm_id
        self.volume_id = volume_id

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        result = await self.netapp_client.create_qtree(self.svm_id, self.volume_id, qspath.name)
        self.netapp_client.check_job_result(result, [])
        if options is not None:
            await self.update_quota_scope(quota_scope_id, options)

    async def describe_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> Optional[QuotaUsage]:
        qspath = self.mangle_qspath(quota_scope_id)
        if not qspath.exists():
            return None
        return await self.netapp_client.get_quota_report(self.svm_id, self.volume_id, qspath.name)

    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        result = await self.netapp_client.set_quota_rule(
            self.svm_id,
            self.volume_id,
            qspath.name,
            config,
        )
        self.netapp_client.check_job_result(result, [])
        result = await self.netapp_client.enable_quota(self.volume_id)
        self.netapp_client.check_job_result(result, ["5308507"])  # pass if "already on"

    # FIXME: How do we implement unset_quota() for NetApp?
    async def unset_quota(self, quota_scope_id: QuotaScopeID) -> None:
        raise InvalidQuotaScopeError(
            "Unsetting folder limit without removing quota scope is not possible for this backend"
        )

    async def delete_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        if len([p for p in qspath.iterdir() if p.is_dir()]) > 0:
            raise NotEmptyError(quota_scope_id)
        # QTree and quota rule is automatically removed
        # when the corresponding directory is deleted.
        await aiofiles.os.rmdir(qspath)


class XCPFSOpModel(BaseFSOpModel):
    """
    Accelerates filesystem operations using NetApp's XCP tool.
    If the tool is not installed or its license is not available, it will gracefully fallback to the
    BaseFSOpModel's standard implementations.

    ref) https://docs.netapp.com/us-en/xcp/xcp-install-xcp.html#install-and-configure-workflow
    """

    def __init__(
        self,
        mount_path: Path,
        scandir_limit: int,
        netapp_nfs_host: str,
        netapp_xcp_cmd: str,
        nas_path: Path,
    ) -> None:
        super().__init__(mount_path, scandir_limit)
        self.netapp_nfs_host = netapp_nfs_host
        self.netapp_xcp_cmd = netapp_xcp_cmd
        self.nas_path = nas_path

    async def check_license(self) -> bool:
        if xcp_lic_check_path.exists() and xcp_lic_check_path.stat().st_mtime >= time.time() - 3600:
            return xcp_lic_check_path.read_bytes().strip() == b"1"
        try:
            proc = await asyncio.create_subprocess_exec(
                *[*self.netapp_xcp_cmd, b"license"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            stdout, _ = await proc.communicate()
        except FileNotFoundError:
            return False
        result = b"License status: ACTIVE\n" in stdout
        xcp_lic_check_path.parent.mkdir(parents=True, exist_ok=True)
        if result:
            xcp_lic_check_path.write_bytes(b"1")
        else:
            xcp_lic_check_path.write_bytes(b"0")
        return result

    async def copy_tree(
        self,
        src_path: Path,
        dst_path: Path,
    ) -> None:
        if not src_path.is_relative_to(self.mount_path):
            raise ValueError(f"Invalid path inside the volume: {src_path}")
        if not dst_path.is_relative_to(self.mount_path):
            raise ValueError(f"Invalid path inside the volume: {dst_path}")

        # Rearrange the paths into the NFS absolute path.
        # These relative paths contains the qtree (quota-scope) name as the first part.
        src_relpath = src_path.relative_to(self.mount_path)
        dst_relpath = dst_path.relative_to(self.mount_path)
        src_nfspath = f"{self.netapp_nfs_host}:{self.nas_path}/{src_relpath}"
        dst_nfspath = f"{self.netapp_nfs_host}:{self.nas_path}/{dst_relpath}"

        copy_cmd = [
            *self.netapp_xcp_cmd,
            b"copy",
            os.fsencode(src_nfspath),
            os.fsencode(dst_nfspath),
        ]
        async with aclosing(spawn_and_watch(copy_cmd)) as ag:
            async for line in ag:
                # TODO: line for bgtask
                pass

    async def delete_tree(self, path: Path) -> None:
        relpath = path.relative_to(self.mount_path)
        nfspath = f"{self.netapp_nfs_host}:{self.nas_path}/{relpath}"
        delete_cmd = [
            *self.netapp_xcp_cmd,
            b"delete",
            b"-force",
            b"-removetopdir",
            os.fsencode(nfspath),
        ]
        async with aclosing(spawn_and_watch(delete_cmd)) as ag:
            async for line in ag:
                # TODO: line for bgtask
                pass

    def scan_tree(
        self,
        target_path: Path,
        recursive=False,
    ) -> AsyncIterator[DirEntry]:
        target_relpath = target_path.relative_to(self.mount_path)
        nfspath = f"{self.netapp_nfs_host}:{self.nas_path}/{target_relpath}"
        # Use a custom formatting
        scan_cmd = [
            *self.netapp_xcp_cmd,
            b"scan",
            b"-fmt",
            rb"'{}\0{}\0{}\0{}\0{}\0{}\0{}\0{}'.format(mode,uid,gid,ctime,mtime,type,size,x)",
            os.fsencode(nfspath),
        ]

        async def aiter() -> AsyncIterator[DirEntry]:
            proc = await asyncio.create_subprocess_exec(
                *scan_cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stderr_lines: list[bytes] = []
            entry_queue: asyncio.Queue[DirEntry | Sentinel] = asyncio.Queue(maxsize=1024)

            async def read_stdout() -> None:
                assert proc.stdout is not None
                while True:
                    line = await proc.stdout.readline()
                    if not line:
                        await entry_queue.put(SENTINEL)
                        break
                    if b"\0" not in line:
                        # This is some unexpected output like
                        # "Error while sending previously unsent statistics"
                        continue
                    parts = tuple(map(os.fsdecode, line.rstrip(b"\n").split(b"\0")))
                    item_path = Path(*Path(parts[7]).parts[2:])
                    inner_relpath = item_path.relative_to(target_relpath)
                    if inner_relpath == Path("."):  # exclude the top dir
                        continue
                    item_abspath = self.mount_path / target_relpath / inner_relpath
                    match int(parts[5]):
                        case 2:
                            entry_type = DirEntryType.DIRECTORY
                        case 5:
                            entry_type = DirEntryType.SYMLINK
                        case _:
                            entry_type = DirEntryType.FILE
                    symlink_target = ""
                    if entry_type == DirEntryType.SYMLINK:
                        symlink_dst = Path(item_abspath).resolve()
                        try:
                            symlink_dst = symlink_dst.relative_to(target_path)
                        except ValueError:
                            pass
                        symlink_target = os.fsdecode(symlink_dst)
                    await entry_queue.put(
                        DirEntry(
                            name=item_path.name,
                            path=inner_relpath,
                            type=entry_type,
                            stat=Stat(
                                size=int(parts[6]),
                                owner=parts[1],
                                mode=int(parts[0]),
                                modified=fstime2datetime(float(parts[4])),
                                created=fstime2datetime(float(parts[3])),
                            ),
                            symlink_target=symlink_target,
                        ),
                    )

            async def read_stderr() -> None:
                assert proc.stderr is not None
                while True:
                    line = await proc.stderr.readline()
                    if not line:
                        break
                    stderr_lines.append(line)

            try:
                async with asyncio.TaskGroup() as tg:
                    tg.create_task(read_stdout())
                    tg.create_task(read_stderr())
                    while True:
                        item = await entry_queue.get()
                        try:
                            if item is SENTINEL:
                                break
                            yield item
                        finally:
                            entry_queue.task_done()
            finally:
                await proc.wait()
            if proc.returncode != 0:
                error_msg_prefix = b"xcp: ERROR: "
                error_msg = "unknown"
                for line in stderr_lines:
                    if line.startswith(error_msg_prefix):
                        error_msg = line.removeprefix(error_msg_prefix).decode()
                        break
                raise ExecutionError(f"Running XCP has failed: {error_msg}")

        return aiter()

    async def scan_tree_usage(
        self,
        target_path: Path,
    ) -> TreeUsage:
        target_relpath = target_path.relative_to(self.mount_path)
        nfspath = f"{self.netapp_nfs_host}:{self.nas_path}/{target_relpath}"
        total_size = 0
        total_count = 0
        # Use a tree statistics output formatting
        scan_cmd = [*self.netapp_xcp_cmd, b"scan", b"-csv", os.fsencode(nfspath)]
        proc = await asyncio.create_subprocess_exec(
            *scan_cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        try:
            async with asyncio.timeout(30):
                stdout, stderr = await proc.communicate()
                if proc.returncode == 0:
                    reader = csv.reader(map(lambda b: b.decode(), stdout.splitlines()))
                    for row in reader:
                        if len(row) < 2:
                            continue
                        match row[0].lower():
                            case "total count":
                                total_count = int(row[1])
                            case "total space used":
                                total_size = int(row[1])
                            case _:
                                pass
                else:
                    error_msg_prefix = b"xcp: ERROR: "
                    error_msg = "unknown"
                    for line in stderr.splitlines():
                        if line.startswith(error_msg_prefix):
                            error_msg = line.removeprefix(error_msg_prefix).rstrip().decode()
                            break
                    raise ExecutionError(f"Running XCP has failed: {error_msg}")
        except asyncio.TimeoutError:
            # -1 indicates "too many"
            total_size = -1
            total_count = -1
        finally:
            if proc.returncode is None:
                proc.kill()
                await proc.wait()
        return TreeUsage(file_count=total_count, used_bytes=total_size)

    async def scan_tree_size(
        self,
        target_path: Path,
    ) -> BinarySize:
        usage = await self.scan_tree_usage(target_path)
        return BinarySize(usage.used_bytes)


class NetAppVolume(BaseVolume):
    name = "netapp"
    ontap_endpoint: str
    netapp_user: str
    netapp_password: str
    svm_name: str
    svm_id: StorageID
    volume_name: str
    volume_id: VolumeID
    nas_path: Path

    async def create_quota_model(self) -> AbstractQuotaModel:
        return QTreeQuotaModel(
            self.mount_path,
            self.netapp_client,
            self.svm_id,
            self.volume_id,
        )

    async def create_fsop_model(self) -> AbstractFSOpModel:
        xcp_fsop_model = XCPFSOpModel(
            self.mount_path,
            self.local_config["storage-proxy"]["scandir-limit"],
            self.netapp_nfs_host,
            self.netapp_xcp_cmd,
            self.nas_path,
        )
        if await xcp_fsop_model.check_license():
            return xcp_fsop_model
        log.warning(
            "XCP is not installed ('{}') or its license is not active. "
            "Falling back to BaseFSOpModel which may be slower.",
            shlex.join(self.netapp_xcp_cmd),
        )
        return BaseFSOpModel(
            self.mount_path,
            self.local_config["storage-proxy"]["scandir-limit"],
        )

    async def init(self) -> None:
        self.ontap_endpoint = self.config["netapp_ontap_endpoint"]
        self.netapp_client = NetAppClient(
            self.ontap_endpoint,
            self.config["netapp_ontap_user"],
            self.config["netapp_ontap_password"],
        )
        self.netapp_nfs_host = self.config["netapp_nfs_host"]
        self.netapp_xcp_cmd = self.config["netapp_xcp_cmd"]
        self.volume_name = self.config["netapp_volume_name"]
        volume_info = await self.netapp_client.get_volume_by_name(self.volume_name, ["svm"])
        assert "svm" in volume_info
        self.volume_id = volume_info["uuid"]
        self.svm_name = volume_info["svm"]["name"]
        self.svm_id = StorageID(volume_info["svm"]["uuid"])
        self.nas_path = volume_info["path"]
        assert self.nas_path.is_absolute()
        # Example volume ID: 8a5c9938-a872-11ed-8519-d039ea42b802
        # Example volume name: "cj1nipacjssd1_02R10c1v2"
        # Example volume path: /cj1nipacjssd1_02R10c1v2/
        # Example qtree ID: 3 (default: 0)
        # Example qtree name: "quotascope1" (default: "")
        # Example qtree path: /cj1nipacjssd1_02R10c1v2/quotascope1
        # nfspath:    192.168.1.7:/cj1nipacjssd1_02R10c1v2/quotascope1/vfid[0:2]/vfid[2:4]/vfid[4:]
        # xcp_host:   ^^^^^^^^^^^
        # volume_path:            ^^^^^^^^^^^^^^^^^^^^^^^^^
        # vfpath: = mangle_vfpath(vfid)     /vfroot/mydata/quotascope1/vfid[0:2]/vfid[2:4]/vfid[4:]
        # vfroot:                           ^^^^^^^
        # mount_path:                       ^^^^^^^^^^^^^^
        # quota-scope path:                 ^^^^^^^^^^^^^^^^^^^^^^^^^^
        # vf_relpath: = vfpath.relative_to(mount_path)     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        # NOTE: QTree ID and name are per-volume.
        #       (i.e., Different volumes may have the same qtree ID and names
        #       for different QTree instances!)
        # NOTE: The default qtree in the volume has no explicit path.
        await super().init()

    async def shutdown(self) -> None:
        await self.netapp_client.aclose()

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset([CAP_VFOLDER, CAP_FAST_FS_SIZE, CAP_FAST_SIZE, CAP_QUOTA, CAP_METRIC])

    async def get_hwinfo(self) -> HardwareMetadata:
        volume_info = await self.netapp_client.get_volume_by_id(self.volume_id, ["files", "quota"])
        metadata = {
            "quota": volume_info["quota"],  # type: ignore
            "files": volume_info["files"],  # type: ignore
        }
        return {"status": "healthy", "status_info": None, "metadata": metadata}

    async def get_fs_usage(self) -> CapacityUsage:
        volume_info = await self.netapp_client.get_volume_by_id(
            self.volume_id, ["space.size,space.used"]
        )
        assert "space" in volume_info
        return CapacityUsage(
            capacity_bytes=BinarySize(volume_info["space"]["size"]),
            used_bytes=BinarySize(volume_info["space"]["used"]),
        )

    async def get_performance_metric(self) -> FSPerfMetric:
        volume_info = await self.netapp_client.get_volume_by_id(self.volume_id, ["statistics"])
        assert "statistics" in volume_info
        stats = volume_info["statistics"]
        # Example of volume info's statistics field:
        # 'statistics': {'iops_raw': {'other': 10860,
        #                             'read': 0,
        #                             'total': 10860,
        #                             'write': 0},
        #                'status': 'ok',
        #                'throughput_raw': {'other': 1239476,
        #                                   'read': 0,
        #                                   'total': 1239476,
        #                                   'write': 0},
        #                'timestamp': '2023-04-04T08:33:04Z'},
        # Example of volume metrics (default: 15s interval of last hour):
        #  {'num_records': 240,
        #   'records': [{'iops': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'latency': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'throughput': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'timestamp': '2023-04-04T08:48:30Z'},
        #               {'iops': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'latency': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'throughput': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'timestamp': '2023-04-04T08:48:15Z'},
        #               {'iops': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'latency': {'other': 66, 'read': 0, 'total': 66, 'write': 0},
        #                'throughput': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'timestamp': '2023-04-04T08:48:00Z'},
        #               {'iops': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'latency': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'throughput': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'timestamp': '2023-04-04T08:47:45Z'},
        #               {'iops': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'latency': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'throughput': {'other': 0, 'read': 0, 'total': 0, 'write': 0},
        #                'timestamp': '2023-04-04T08:47:30Z'},
        #               ...
        return FSPerfMetric(
            iops_read=stats["iops_raw"]["read"],
            iops_write=stats["iops_raw"]["write"],
            io_bytes_read=stats["throughput_raw"]["read"],
            io_bytes_write=stats["throughput_raw"]["write"],
            io_usec_read=0,
            io_usec_write=0,
        )



from typing import Optional


class DellError(Exception):
    pass


class DellInitError(DellError):
    pass


class DellAPIError(DellError):
    message: str

    def __init__(self, message: Optional[str] = None, *args):
        super().__init__(message, *args)

        self.message = message or ""

    def __str__(self) -> str:
        return type(self).__name__


class DellInvalidBodyError(DellAPIError):
    pass


class DellUnauthorizedError(DellAPIError):
    pass


class DellNotFoundError(DellAPIError):
    pass


class DellInternalError(DellAPIError):
    pass


class DellNoMetricError(DellError):
    pass



import asyncio
import logging
import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import (
    Any,
    Dict,
    FrozenSet,
    List,
    Optional,
)

import aiofiles
import aiofiles.os

from ai.backend.common.lock import FileLock
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import QuotaScopeID
from ai.backend.storage.abc import CAP_QUOTA, CAP_VFOLDER

from ..abc import AbstractQuotaModel
from ..exception import InvalidQuotaScopeError, NotEmptyError
from ..subproc import run
from ..types import (
    QuotaConfig,
    QuotaUsage,
)
from ..vfs import BaseQuotaModel, BaseVolume

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

LOCK_FILE = Path("/tmp/backendai-xfs-file-lock")
Path(LOCK_FILE).touch()


class XfsProjectRegistry:
    file_projects: Path = Path("/etc/projects")
    file_projid: Path = Path("/etc/projid")
    backend: BaseVolume
    name_id_map: Dict[str, int] = dict()
    project_id_pool: List[int] = list()

    async def init(self, backend: BaseVolume) -> None:
        self.backend = backend

    async def read_project_info(self):
        def _read_projid_file():
            return self.file_projid.read_text()

        # TODO: how to handle if /etc/proj* files are deleted by external reason?
        # TODO: do we need to use /etc/proj* files to enlist the project information?
        if self.file_projid.is_file():
            project_id_pool = []
            self.name_id_map = {}
            loop = asyncio.get_running_loop()
            raw_projid = await loop.run_in_executor(None, _read_projid_file)
            for line in raw_projid.splitlines():
                proj_name, proj_id = line.split(":")[:2]
                project_id_pool.append(int(proj_id))
                self.name_id_map[proj_name] = int(proj_id)
            self.project_id_pool = sorted(project_id_pool)
        else:
            await run(["sudo", "touch", self.file_projid])
        if not Path(self.file_projects).is_file():
            await run(["sudo", "touch", self.file_projects])

    async def add_project_entry(
        self,
        quota_scope_id: QuotaScopeID,
        qspath: Path,
        *,
        project_id: Optional[int] = None,
    ) -> None:
        if project_id is None:
            project_id = self.get_free_project_id()

        temp_name_projects = ""
        temp_name_projid = ""

        def _create_temp_files():
            nonlocal temp_name_projects, temp_name_projid
            _tmp_projects = NamedTemporaryFile(delete=False)
            _tmp_projid = NamedTemporaryFile(delete=False)
            try:
                _projects_content = Path(self.file_projects).read_text()
                if _projects_content.strip() != "" and not _projects_content.endswith(
                    "\n",
                ):
                    _projects_content += "\n"
                _projects_content += f"{project_id}:{qspath}\n"
                _tmp_projects.write(_projects_content.encode("ascii"))
                temp_name_projects = _tmp_projects.name

                _projid_content = Path(self.file_projid).read_text()
                if _projid_content.strip() != "" and not _projid_content.endswith("\n"):
                    _projid_content += "\n"
                _projid_content += f"{quota_scope_id.pathname}:{project_id}\n"
                _tmp_projid.write(_projid_content.encode("ascii"))
                temp_name_projid = _tmp_projid.name
            finally:
                _tmp_projects.close()
                _tmp_projid.close()

        def _delete_temp_files():
            try:
                os.unlink(temp_name_projects)
            except FileNotFoundError:
                pass
            try:
                os.unlink(temp_name_projid)
            except FileNotFoundError:
                pass

        loop = asyncio.get_running_loop()
        try:
            await loop.run_in_executor(None, _create_temp_files)
            await run(["sudo", "cp", "-rp", temp_name_projects, self.file_projects])
            await run(["sudo", "cp", "-rp", temp_name_projid, self.file_projid])
        finally:
            await loop.run_in_executor(None, _delete_temp_files)

    async def remove_project_entry(self, quota_scope_id: QuotaScopeID) -> None:
        await run(["sudo", "sed", "-i.bak", f"/{quota_scope_id.pathname}/d", self.file_projects])
        await run(["sudo", "sed", "-i.bak", f"/{quota_scope_id.pathname}/d", self.file_projid])

    def get_free_project_id(self) -> int:
        """
        Get the next project_id, which is the smallest unused integer.
        """
        project_id = -1
        for i in range(len(self.project_id_pool) - 1):
            if self.project_id_pool[i] + 1 != self.project_id_pool[i + 1]:
                project_id = self.project_id_pool[i] + 1
                break
        if len(self.project_id_pool) == 0:
            project_id = 1
        if project_id == -1:
            project_id = self.project_id_pool[-1] + 1
        return project_id


class XFSProjectQuotaModel(BaseQuotaModel):
    """
    Implements the quota scope model using XFS projects.
    """

    def __init__(
        self,
        mount_path: Path,
        project_registry: XfsProjectRegistry,
    ) -> None:
        super().__init__(mount_path)
        self.project_registry = project_registry
        stat_vfs = os.statvfs(mount_path)
        self.block_size = stat_vfs.f_bsize

    async def create_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        options: Optional[QuotaConfig] = None,
        extra_args: Optional[dict[str, Any]] = None,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        try:
            if options is None:
                # Set the limit as the filesystem size
                vfs_stat = os.statvfs(self.mount_path)
                options = QuotaConfig(vfs_stat.f_blocks * self.block_size)
            async with FileLock(LOCK_FILE):
                log.info(
                    "creating project quota (qs:{}, q:{})",
                    quota_scope_id,
                    (options.limit_bytes if options else None),
                )
                await aiofiles.os.makedirs(qspath)
                await self.project_registry.read_project_info()
                await self.project_registry.add_project_entry(quota_scope_id, qspath)
                await self.project_registry.read_project_info()
        except (asyncio.CancelledError, asyncio.TimeoutError):
            log.exception("quota-scope creation timeout")
            raise
        except Exception:
            log.exception("quota-scope creation error")
            raise
        if options is not None:
            await self.update_quota_scope(quota_scope_id, options)

    async def describe_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> Optional[QuotaUsage]:
        if not self.mangle_qspath(quota_scope_id).exists():
            return None
        full_report = await run(
            # -p: project quota only
            # -b: as number of blocks
            # -N: without header
            ["sudo", "xfs_quota", "-x", "-c", "report -p -b -N", self.mount_path],
        )
        print(full_report)
        for line in full_report.splitlines():
            if quota_scope_id.pathname in line:
                report = line
                break
        else:
            raise RuntimeError(f"unknown xfs project ID: {quota_scope_id.pathname}")
        if len(report.split()) != 6:
            raise ValueError("unexpected format for xfs_quota report")
        _, used_kbs, _, hard_limit_kbs, _, _ = report.split()
        # By default, report command displays the sizes in the 1 KiB unit.
        used_bytes = int(used_kbs) * 1024
        hard_limit_bytes = int(hard_limit_kbs) * 1024
        return QuotaUsage(used_bytes, hard_limit_bytes)

    async def update_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
        config: QuotaConfig,
    ) -> None:
        # This will annotate all entries under the quota scope tree as a part of the project.
        await run(
            [
                "sudo",
                "xfs_quota",
                "-x",
                "-c",
                f"project -s {quota_scope_id.pathname}",
                self.mount_path,
            ],
        )
        # bsoft, bhard accepts bytes or binary-prefixed numbers.
        await run(
            [
                "sudo",
                "xfs_quota",
                "-x",
                "-c",
                (
                    "limit -p"
                    f" bsoft={config.limit_bytes} bhard={config.limit_bytes} {quota_scope_id.pathname}"
                ),
                self.mount_path,
            ],
        )

    async def unset_quota(self, quota_scope_id: QuotaScopeID) -> None:
        raise InvalidQuotaScopeError(
            "Unsetting folder limit without removing quota scope is not possible for this backend"
        )

    async def delete_quota_scope(
        self,
        quota_scope_id: QuotaScopeID,
    ) -> None:
        qspath = self.mangle_qspath(quota_scope_id)
        if len([p for p in qspath.iterdir() if p.is_dir()]) > 0:
            raise NotEmptyError(quota_scope_id)
        async with FileLock(LOCK_FILE):
            await self.project_registry.read_project_info()
            await self.project_registry.remove_project_entry(quota_scope_id)
            await self.project_registry.read_project_info()
            await aiofiles.os.rmdir(qspath)


class XfsVolume(BaseVolume):
    """
    XFS volume backend. XFS natively supports per-directory quota through
    the project qutoa. To enalbe project quota, the XFS volume should be
    mounted with `-o pquota` option.

    This backend requires `root` or no password `sudo` permission to run
    `xfs_quota` command and write to `/etc/projects` and `/etc/projid`.
    """

    name = "xfs"

    project_registry: XfsProjectRegistry

    async def init(self) -> None:
        self.project_registry = XfsProjectRegistry()
        await self.project_registry.init(self)
        await super().init()

    async def create_quota_model(self) -> AbstractQuotaModel:
        return XFSProjectQuotaModel(
            self.mount_path,
            self.project_registry,
        )

    async def get_capabilities(self) -> FrozenSet[str]:
        return frozenset([CAP_VFOLDER, CAP_QUOTA])
